# -*- coding: utf-8 -*-

# Copyright 2016 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, unused-argument

"""
yabt C++ Builder
~~~~~~~~~~~~~~~~

:author: Itamar Ostricher

TODO: support injecting compile/link flags for 3rd party libs
TODO: CppSharedLib builder
"""


from hashlib import md5
from os.path import basename, dirname, join, relpath, splitext

from ostrich.utils.collections import listify

from yabt.docker import extend_runtime_params, format_docker_run_params
from ..artifact import ArtifactType as AT
from .dockerapp import build_app_docker_and_bin, register_app_builder_sig
from ..extend import (
    PropType as PT, register_build_func, register_builder_sig,
    register_manipulate_target_hook, register_test_func,
    register_cache_json_func)
from ..logging import make_logger
from ..target_utils import split, Target
from ..utils import link_files, rmtree, yprint


logger = make_logger(__name__)


class CompilerConfig:
    """Helper class for managing compiler / linker options and flags.

    Design note:
    Notice the `extra_{compile,link}_flags` for target X are collected from
    all the dependencies of target X, but are actually applied to a command
    that is executed within the **build-env image** of target X, which can
    have a completely unrelated subgraph!
    It is the responsibility of the "user" to use these parameters correctly,
    and make sure that the build-env-image contains the **build-time**
    dependencies that are needed to make this parameters work *there*, while
    the target itself has the correct **runtime** dependencies.
    See an example of how this plays out in the YaBT tests - specifically,
    check out the `cpp/hello_boost` test and examine how the build-env image
    contains boost dev-lib (for build-time, including all boost sub-libs),
    while the final target depends on boost-runtime, which installs only 3
    boost shared objects that are needed for this specific application.

    TODO: add this note to proper project docs (also - create proper docs...)
    """

    def __init__(self, build_context, target, extra_params=None):
        self.compiler = self.get(
            'compiler', build_context.conf, target, 'g++')
        self.linker = self.get(
            'linker', build_context.conf, target, 'g++')
        self.compile_flags = list(self.get(
            'compile_flags', build_context.conf, target, []))
        self.link_flags = list(self.get(
            'link_flags', build_context.conf, target, []))
        self.include_path = list(self.get(
            'include_path', build_context.conf, target, []))

        def generate_extra_params():
            if extra_params:
                yield extra_params
            yield target.props.build_params
            # using topological order here because linker `-l<lib>` flags
            # are sensitive to the order that they appear in!
            # so if this target depends on, for example, both libsoft and
            # libfftw, and also libsoft requires symbols that are defined in
            # libfftw, then `-lfftw` must appear *after* `-lsoft`.
            yield from (
                dep.props.build_params for dep in
                build_context.walk_target_deps_topological_order(target))

        for build_params in reversed(list(generate_extra_params())):
            self.compile_flags.extend(
                listify(build_params.get('extra_compile_flags')))
            self.link_flags.extend(
                listify(build_params.get('extra_link_flags')))

    def as_dict(self):
        return {key: getattr(self, key)
                for key in ('compiler', 'linker', 'compile_flags',
                            'link_flags', 'include_path')}

    def get(self, param, config, target, fallback):
        """Return the value of `param`, according to priority / expansion.

        First priority - the target itself.
        Second priority - the project config.
        Third priority - a global default ("fallback").

        In list-params, a '$*' term processed as "expansion term", meaning
        it is replaced with all terms from the config-level.
        """
        target_val = target.props.get(param)
        config_val = config.get(param, fallback)
        if not target_val:
            return config_val
        if isinstance(target_val, list):
            val = []
            for el in target_val:
                if el == '$*':
                    val.extend(listify(config_val))
                else:
                    val.append(el)
            return val
        return target_val


# Common C++ builder signature terms
CPP_SIG = [
     ('sources', PT.FileList),
     ('in_buildenv', PT.Target),
     ('headers', PT.FileList, None),
     ('protos', PT.TargetList, None),
     ('cmd_env', None),
     ('compiler', PT.str, None),
     ('linker', PT.str, None),
     ('compile_flags', PT.StrList, None),
     ('link_flags', PT.StrList, None),
     ('include_path', PT.StrList, None),
     # an internal "prop" used internally bt YBT to add compiler config data
     # to the target props, so it is considered by target hashing (for cache)
     ('_internal_dict_', PT.dict, None),
]


register_app_builder_sig(
    'CppApp', [('executable', PT.File, None), ('main', PT.Target, None)])


@register_manipulate_target_hook('CppApp')
def cpp_app_manipulate_target(build_context, target):
    logger.debug('Injecting {} to deps of {}',
                 target.props.base_image, target.name)
    target.deps.append(target.props.base_image)
    if target.props.main and target.props.main not in target.deps:
        logger.debug('Injecting {} to deps of {}',
                     target.props.main, target.name)
        target.deps.append(target.props.main)


@register_build_func('CppApp')
def cpp_app_builder(build_context, target):
    """Pack a C++ binary as a Docker image with its runtime dependencies.

    TODO(itamar): Dynamically analyze the binary and copy shared objects
    from its buildenv image to the runtime image, unless they're installed.
    """
    yprint(build_context.conf, 'Build CppApp', target)
    if target.props.executable and target.props.main:
        raise KeyError(
            '`main` and `executable` arguments are mutually exclusive')
    if target.props.executable:
        if target.props.executable not in target.artifacts.get(AT.app):
            target.artifacts.add(AT.app, target.props.executable)
        entrypoint = [target.props.executable]
    elif target.props.main:
        prog = build_context.targets[target.props.main]
        binary = list(prog.artifacts.get(AT.binary).keys())[0]
        entrypoint = ['/usr/src/bin/' + binary]
    else:
        raise KeyError('Must specify either `main` or `executable` argument')
    build_app_docker_and_bin(
        build_context, target, entrypoint=entrypoint)


register_builder_sig('CppProg', CPP_SIG)
register_builder_sig(
    'CppGTest',
    CPP_SIG + [
        ('test_flags', PT.StrList, None),  # flags to append to test command
        ('test_env', None),  # env vars to inject in test process
        # TODO: support different testenv image for test execution
        # ('in_testenv', PT.Target, None),
    ]
)


def make_pre_build_hook(extra_compiler_config_params):
    """Return a pre-build hook function for C++ builders.

    When called, during graph build, it computes and stores the compiler-config
    object on the target, as well as adding it to the internal_dict prop for
    hashing purposes.
    """

    def pre_build_hook(build_context, target):
        target.compiler_config = CompilerConfig(
            build_context, target, extra_compiler_config_params)
        target.props._internal_dict_['compiler_config'] = (
            target.compiler_config.as_dict())

    return pre_build_hook


@register_manipulate_target_hook('CppProg')
def cpp_prog_manipulate_target(build_context, target):
    target.buildenv = target.props.in_buildenv
    target.pre_build_hook = make_pre_build_hook({})


@register_manipulate_target_hook('CppGTest')
def cpp_gtest_manipulate_target(build_context, target):
    target.tags.add('testable')
    target.buildenv = target.props.in_buildenv
    # target.buildenvs.append(target.props.in_testenv)
    # manipulate the test_flags prop during target extraction (as opposed to
    # during build func), so it is considered during target hashing (for cache)
    target.props.test_flags.extend(listify(
        build_context.conf.get('gtest_params', {}).get('extra_exec_flags')))
    target.pre_build_hook = make_pre_build_hook(
        build_context.conf.get('gtest_params', {}))


def is_cc_file(filename: str) -> bool:
    return splitext(filename)[-1].lower() in ('.cc', '.cpp', '.cxx', '.c++')


def is_h_file(filename: str) -> bool:
    return splitext(filename)[-1].lower() in ('.h', '.hpp', '.hh', '.hxx')


def compile_cc(build_context, compiler_config, buildenv, sources,
               workspace_dir, buildenv_workspace, cmd_env):
    """Compile list of C++ source files in a buildenv image
       and return list of generated object file.
    """
    objects = []
    for src in sources:
        obj_rel_path = '{}.o'.format(splitext(src)[0])
        obj_file = join(buildenv_workspace, obj_rel_path)
        include_paths = [buildenv_workspace] + compiler_config.include_path
        compile_cmd = (
            [compiler_config.compiler, '-o', obj_file, '-c'] +
            compiler_config.compile_flags +
            ['-I{}'.format(path) for path in include_paths] +
            [join(buildenv_workspace, src)])
        # TODO: capture and transform error messages from compiler so file
        # paths match host paths for smooth(er) editor / IDE integration
        build_context.run_in_buildenv(buildenv, compile_cmd, cmd_env)
        objects.append(
            join(relpath(workspace_dir, build_context.conf.project_root),
                 obj_rel_path))
    return objects


def link_cpp_artifacts(build_context, target, workspace_dir,
                       include_objects: bool):
    """Link required artifacts from dependencies under target workspace dir.
       Return list of object files of dependencies (if `include_objects`).

    Includes:
    - Generated code from proto dependencies
    - Header files from all dependencies
    - Generated header files from all dependencies
    - If `include_objects` is True, also object files from all dependencies
      (these will be returned without linking)
    """
    # include the source & header files of the current target
    # add objects of all dependencies (direct & transitive), if needed
    source_files = target.props.sources + target.props.headers
    generated_srcs = {}
    objects = []

    # add headers of dependencies
    for dep in build_context.generate_all_deps(target):
        source_files.extend(dep.props.get('headers', []))

    link_files(source_files, workspace_dir, None, build_context.conf)

    # add generated headers and collect objects of dependencies
    for dep in build_context.generate_all_deps(target):
        dep.artifacts.link_types(workspace_dir, [AT.gen_h], build_context.conf)
        if include_objects:
            objects.extend(dep.artifacts.get(AT.object).values())

    # add generated code from proto dependencies
    for proto_dep_name in target.props.protos:
        proto_dep = build_context.targets[proto_dep_name]
        proto_dep.artifacts.link_types(workspace_dir, [AT.gen_cc],
                                       build_context.conf)

    return objects


def get_source_files(target, build_context) -> list:
    """Return list of source files for `target`."""
    all_sources = list(target.props.sources)
    for proto_dep_name in target.props.protos:
        proto_dep = build_context.targets[proto_dep_name]
        all_sources.extend(proto_dep.artifacts.get(AT.gen_cc).keys())
    return all_sources


def build_cpp(build_context, target, compiler_config, workspace_dir):
    """Compile and link a C++ binary for `target`."""
    rmtree(workspace_dir)
    binary = join(*split(target.name))
    objects = link_cpp_artifacts(build_context, target, workspace_dir, True)
    buildenv_workspace = build_context.conf.host_to_buildenv_path(
        workspace_dir)
    objects.extend(compile_cc(
        build_context, compiler_config, target.props.in_buildenv,
        get_source_files(target, build_context), workspace_dir,
        buildenv_workspace, target.props.cmd_env))
    bin_file = join(buildenv_workspace, binary)
    link_cmd = (
        [compiler_config.linker, '-o', bin_file] +
        objects + compiler_config.link_flags)
    build_context.run_in_buildenv(
        target.props.in_buildenv, link_cmd, target.props.cmd_env)
    target.artifacts.add(AT.binary, relpath(join(workspace_dir, binary),
                         build_context.conf.project_root), binary)


@register_build_func('CppProg')
def cpp_prog_builder(build_context, target):
    """Build a C++ binary executable"""
    yprint(build_context.conf, 'Build CppProg', target)
    workspace_dir = build_context.get_workspace('CppProg', target.name)
    build_cpp(build_context, target, target.compiler_config, workspace_dir)

    # TODO: remove?
    # # Copy binary artifacts to external destination
    # if target.props.copy_bin_to:
    #     link_artifacts([join(workspace_dir, binary)],
    #                    target.props.copy_bin_to,
    #                    workspace_dir, build_context.conf)


@register_build_func('CppGTest')
def cpp_gtest_builder(build_context, target):
    """Build a C++ test executable"""
    yprint(build_context.conf, 'Build CppGTest', target)
    workspace_dir = build_context.get_workspace('CppGTest', target.name)
    build_cpp(build_context, target, target.compiler_config, workspace_dir)


@register_test_func('CppGTest')
def cpp_gtest_tester(build_context, target):
    """Run a C++ test executable"""
    yprint(build_context.conf, 'Run CppGTest', target)
    workspace_dir = build_context.get_workspace('CppGTest', target.name)
    buildenv_workspace = build_context.conf.host_to_buildenv_path(
        workspace_dir)
    test_cmd = [join(buildenv_workspace, *split(target.name))]
    # take gtest exec flags from the target & from project config
    test_cmd.extend(target.props.test_flags)
    run_params = extend_runtime_params(
                target.props.runtime_params,
                build_context.walk_target_deps_topological_order(target),
                build_context.conf.runtime_params, True)
    build_context.run_in_buildenv(
        # TODO: target.props.in_testenv,
        target.props.in_buildenv, test_cmd, target.props.test_env,
        run_params=format_docker_run_params(run_params))

register_builder_sig('CppLib', CPP_SIG)


@register_manipulate_target_hook('CppLib')
def cpp_lib_manipulate_target(build_context, target):
    target.buildenv = target.props.in_buildenv
    # proto targets are also direct dependencies
    target.deps.extend(target.props.protos)
    target.pre_build_hook = make_pre_build_hook({})


@register_build_func('CppLib')
def cpp_lib_builder(build_context, target):
    """Build C++ object files"""
    yprint(build_context.conf, 'Build CppLib', target)
    workspace_dir = build_context.get_workspace('CppLib', target.name)
    workspace_src_dir = join(workspace_dir, 'src')
    rmtree(workspace_src_dir)
    link_cpp_artifacts(build_context, target, workspace_src_dir, False)
    buildenv_workspace = build_context.conf.host_to_buildenv_path(
        workspace_src_dir)
    objects = compile_cc(
        build_context, target.compiler_config, target.props.in_buildenv,
        get_source_files(target, build_context), workspace_src_dir,
        buildenv_workspace, target.props.cmd_env)
    for obj_file in objects:
        target.artifacts.add(AT.object, obj_file)


def calc_hash(json_str):
    m = md5()
    m.update(json_str.encode('utf8'))
    return m.hexdigest()


def get_deps_specific_hash(build_context, target, dep_type, hash_name):
    hashes = []
    for dep_name in listify(target.deps):
        dep_target = build_context.targets[dep_name]
        if dep_target.builder_name == dep_type:
            if not hasattr(dep_target, hash_name):
                dep_target.compute_json(build_context)
            hashes.append(getattr(dep_target, hash_name))
        else:
            hashes.append(dep_target.hash(build_context))
    return hashes


@register_cache_json_func('CppLib')
def cpp_lib_cache_json(build_context, target: Target):
    """
    When a file `a.cc` includes `b.h`, cpp compiler allows us to not compile
    `a.cc` if only `b.cc` is changed since last compilation. In this case we
    only need to compile  `b.cc` and if `a.cc` contains a main function
    (in our case, it is a CppProg or CppGTest target), link a.

    To support it, we define 3 different hashes:
    - Full hash: props, files, full hashes of CppLib deps, cache hashes
                 (the hash used to access the cache) of all other deps.
    - Headers hash: props, headers files, headers hashes of CppLib deps, cache
                    hashes of all other deps.
    - Sources hash: props, files (headers & sources), headers hashes of CppLib
                    deps, cache hashes of all other deps.

    The hash used to access the cache is the sources hash.
    """
    full_hashes = get_deps_specific_hash(build_context, target, 'CppLib',
                                         '_full_hash')
    headers_hashes = get_deps_specific_hash(build_context, target, 'CppLib',
                                            '_headers_hash')
    full_json = target.compute_target_json(build_context, [], full_hashes)
    headers_json = target.compute_target_json(build_context, ['sources'],
                                              headers_hashes)
    sources_json = target.compute_target_json(build_context, [],
                                              headers_hashes)

    target._full_hash = calc_hash(full_json)
    target._headers_hash = calc_hash(headers_json)
    return sources_json


@register_cache_json_func('CppProg')
def cpp_prog_cache_json(build_context, target: Target):
    """
    We want to link if any CppLib we depend on was changed. So the hash we use
    to access the cache contains props, files, full hashes of CppLib deps
    and cache hashes of all other deps.
    """
    full_hashes = get_deps_specific_hash(build_context, target, 'CppLib',
                                         '_full_hash')
    return target.compute_target_json(build_context, [], full_hashes)


@register_cache_json_func('CppGTest')
def cpp_gtest_cache_json(build_context, target: Target):
    """
    Same as in CppProg
    """
    full_hashes = get_deps_specific_hash(build_context, target, 'CppLib',
                                         '_full_hash')
    return target.compute_target_json(build_context, [], full_hashes)



# -*- coding: utf-8 -*-

# Copyright 2019 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
utilities for tests
~~~~~~~~~~~~~~~~~~~~~~~

:author: Dana Shamir
"""
import re
import random
import string
import networkx as nx
from datetime import datetime


def generate_random_dag(nodes, min_rank=0, max_rank=10, edge_prob=0.3):
    """Return a random DAG with nodes from `nodes`.

    and edges can only go from nodes[i] -> nodes[j] if i < j
    (guaranteeing DAGness).
    """
    g = nx.DiGraph()
    g.add_nodes_from(nodes)
    for j in range(1, len(nodes)):
        rank = random.randint(min_rank, min(j, max_rank))
        g.add_edges_from((nodes[i], nodes[j])
                         for i in random.sample(range(j), k=rank)
                         if random.random() > edge_prob)
    return g


class CBuildTrgtTest:
    """Class for creating a random package
        with projects and their dependencies.

        The package contains several projects. All projects consist of files.
        Single-file projects are also allowed. Some projects may be
        independent of others, some should depend on others, such as
        an installation script. Not all dependencies are valid, for example,
        the library cannot depend on the executable file. All applicable rules
        and dependencies were originally read from a json file. Then I
        hard-coded the contents of the most successful file into sort_test_cfg.
            The CBuildTrgtType class contains the name and properties of the
        project type. In the constructor of the CBuildTrgtTest class, all
        project types are written to the trgs variable and configuration data
        to the min_name_len, max_name_len, steps, max_step_nodes
        variables. File names will be randomly generated from
        min_name_len to max_name_len in length.
            A random package is created in the create_rand_graph function.
        In the loop, we call the __add_nodes function steps times.
        The __add_nodes function adds a maximum of MaxStepNodes
        files. Each file is added along with its random dependencies in
        accordance with the rules from the trgs table. Since files without
        parents are added at each step, the cycle cannot be formed
        on a common graph. Thus, the create_rand_graph function
        always returns a Directed Acyclic Graph (DAG)
    """
    sort_test_cfg = [
        {
            "BuildTargets": [
                {
                    "Name": "AptPackage", "Childs": ["AptPackage"],
                    "MinChilds": 0
                },
                {
                    "Name": "CppGTest", "Childs": ["CppLib", "CppProg"],
                    "MinChilds": 1
                },
                {
                    "Name": "CppLib", "Childs": ["Proto", "AptPackage"],
                    "MinChilds": 1
                },
                {
                    "Name": "CppProg",
                    "Childs": ["CppLib", "Proto", "AptPackage"],
                    "MinChilds": 1
                },
                {
                    "Name": "CustomInstaller", "Childs":
                    [
                        "CppLib", "Proto", "CppProg", "CppGTest",
                        "Python", "PythonPackage", "PythonTest"
                    ], "MinChilds": 3
                },
                {
                    "Name": "Proto", "Childs": ["Proto"], "MinChilds": 0
                },
                {
                    "Name": "Python", "Childs": ["Python", "PythonPackage"],
                    "MinChilds": 0
                },
                {
                    "Name": "PythonPackage", "Childs": ["PythonPackage"],
                    "MinChilds": 0
                },
                {
                    "Name": "PythonTest",
                    "Childs": ["PythonPackage", "Python"],
                    "MinChilds": 2
                }
            ]
        },
        {
            "Config": {
                "MinNameLen": 2, "MaxNameLen": 5,
                "Steps": 100, "MaxStepNodes": 8
            }
        }
    ]

    class CBuildTrgtType:
        """ Type of project"""
        def __init__(self, trg):
            self.name = trg['Name']
            self.childs = trg['Childs']  # Types of available dependencies
            self.min_childs = trg['MinChilds']  # Minimum of dependencies

        def is_child(self, trg):
            return trg.name in self.childs

    def __init__(self):
        bd_trgs = CBuildTrgtTest.sort_test_cfg[0]['BuildTargets']
        self.trgs = list()
        for trg in bd_trgs:
            self.trgs.append(CBuildTrgtTest.CBuildTrgtType(trg))
        cfg = CBuildTrgtTest.sort_test_cfg[1]['Config']
        self.min_name_len = cfg['MinNameLen']
        self.max_name_len = cfg['MaxNameLen']
        self.steps = cfg['Steps']
        self.max_step_nodes = cfg['MaxStepNodes']
        self.letters = string.ascii_lowercase

    def __new_build_trgt(self):
        """Generate a random string """
        str_len = random.randint(self.min_name_len, self.max_name_len)
        name = ''.join(random.choice(self.letters) for i in range(str_len))
        i = random.randint(0, len(self.trgs) - 1)
        return [name, self.trgs[i]]

    def __add_nodes(self, graph):
        """ Added random nodes with its random dependencies"""
        edges = dict()
        nod_num = random.randint(1, self.max_step_nodes)
        for _ in range(nod_num):
            trg = self.__new_build_trgt()
            edges.update({trg[0]: trg[1]})
        nodes = list(graph)
        trg_attrs = nx.get_node_attributes(graph, 'trg')
        nodes_len = len(nodes)
        for name in edges:
            if name not in graph:
                trg = edges[name]
                childs = 0
                if nodes_len > 0:
                    child_num = random.randint(0, nodes_len - 1)
                    for i in range(0, child_num):
                        child_name = nodes[i]
                        child_trg = trg_attrs[child_name]
                        if trg.is_child(child_trg):
                            graph.add_edge(name, child_name)
                            childs = childs + 1
                if trg.min_childs <= childs:
                    graph.add_node(name, trg=trg)
                elif childs > 0:
                    graph.remove_node(name)

    def create_rand_graph(self):
        """ Created string random DAG """
        graph = nx.DiGraph()
        for _ in range(self.steps):
            self.__add_nodes(graph)
        return graph


def write_test_dot(graph, out_fname=''):
    """Write build graph in dot format to `out_f` file-like object."""
    if not out_fname:
        now = datetime.now()
        out_fname = now.strftime('ybt_%2d_%2m_%Y_%2H_%2M_%2S_%f.dot')
    fout = open(out_fname, 'w')
    fout.write('strict digraph    {\n')
    clrs = nx.get_node_attributes(graph, 'color')
    stls = nx.get_node_attributes(graph, 'style')
    fclrs = nx.get_node_attributes(graph, 'fillcolor')
    for node in graph.nodes():
        s = '['
        if node in clrs:
            s = s + 'color=' + str(clrs[node]) + ','
        if node in stls:
            s = s + 'color=' + str(stls[node]) + ','
        if node in fclrs:
            s = s + 'color=' + str(fclrs[node]) + ','
        fout.write('  {} {}];\n'.format(node, s))
    fout.writelines('    {} -> {};\n'.format(u, v)
                    for u, v in graph.edges())
    fout.write('}\n\n')
    fout.close()
    print('Output file is "{}"'.format(out_fname))
    return out_fname


def dot_file_line_proc(graph, line):
    """ Parsing dot file line """
    edge = re.split(r'\s*->\s*', line)
    if len(edge) == 2:
        name = re.sub(r'\s+|;', '', edge[0]).strip('\"')
        child_name = re.sub(r'\s+|;', '', edge[1]).strip('\"')
        graph.add_edge(name, child_name)
    else:
        temp = re.split(r'\s+', line, 2)
        if (not temp[0]) & bool(temp[1]):
            name = temp[1].strip('\"')
            attributes = dict()
            for attr_str in re.split(r'\[|,|\]|;|\n', temp[2]):
                attr = re.split('=', attr_str)
                if len(attr) == 2:
                    attributes.update({
                        attr[0].strip('\"'): attr[1].strip('\"')})
            graph.add_node(name, **attributes)


def load_dot(in_fname):
    """ Read graph from dot format """
    with open(in_fname, 'r') as in_f:
        graph = nx.DiGraph()
        for line in in_f:
            dot_file_line_proc(graph, line)
    return graph



# -*- coding: utf-8 -*-

# Copyright 2016 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt target utils module
~~~~~~~~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""


from collections import defaultdict
from hashlib import md5
import json
from os.path import join, normpath
from pathlib import PurePath
import types

from munch import Munch
from ostrich.utils.collections import listify
from ostrich.utils.text import get_safe_path

from .artifact import ArtifactStore
from .extend import Plugin, PropType as PT
from .compat import walk
from .config import Config
from .utils import hash_tree, norm_proj_path


_TARGET_NAMES_WHITELIST = frozenset(('*', '@default'))


def validate_name(target_name):
    try:
        if (target_name in _TARGET_NAMES_WHITELIST or
                target_name == get_safe_path(target_name)):
            return target_name
    except ValueError:
        pass
    raise ValueError("Invalid target name: `{}'".format(target_name))


def split(target_name):
    """Split a target name. Returns a tuple "(build_module, name)".

    The split is on the first `:`.
    Extra `:` are considered part of the name.
    """
    return target_name.split(':', 1)


def split_build_module(target_name):
    """Return the build module component of a target name."""
    return split(target_name)[0]


def split_name(target_name):
    """Return the name component of a target name."""
    return split(target_name)[1]


def norm_name(build_module: str, target_name: str):
    """Return a normalized canonical target name for the `target_name`
       observed in build module `build_module`.

    A normalized canonical target name is of the form "<build module>:<name>",
    where <build module> is the relative normalized path from the project root
    to the target build module (POSIX), and <name> is a valid target name
    (see `validate_name()`).
    """
    if ':' not in target_name:
        raise ValueError(
            "Must provide fully-qualified target name (with `:') to avoid "
            "possible ambiguity - `{}' not valid".format(target_name))

    mod, name = split(target_name)
    return '{}:{}'.format(
        PurePath(norm_proj_path(mod, build_module)).as_posix().strip('.'),
        validate_name(name))


def expand_target_selector(target_selector: str, conf: Config):
    """Return a normalized target name (where `**:*` is the normalized form of
       itself).

    Target specifier can be:

    - `**:*` - means to recursively build all targets under current
      working dir.
    - relative path from current working directory to another directory -
        means to build all targets defined in that build module.
    - a name of a target - means to build this named target in the build module
        in the current working directory.
    - a named target in another build module, with the build module given as a
        relative path from the current working directory (e.g. `../foo:bar`) -
        means to build the specified named target in the specified build
        module.
    - in cases where a relative path can be specified, it should be given using
        standard POSIX relative path construction.
    """
    if target_selector == '**:*':
        return target_selector
    if ':' not in target_selector:
        target_selector += ':*'
    build_module, target_name = split(target_selector)
    build_module = normpath(join(conf.get_rel_work_dir(), build_module))
    return '{}:{}'.format(PurePath(build_module).as_posix().strip('.'),
                          validate_name(target_name))


def parse_target_selectors(target_selectors: list, conf: Config):
    return [expand_target_selector(target_selector, conf)
            for target_selector in target_selectors]


def hashify_targets(targets: list, build_context) -> list:
    """Return sorted hashes of `targets`."""
    return sorted(build_context.targets[target_name].hash(build_context)
                  for target_name in listify(targets))


def hashify_files(files: list) -> dict:
    """Return mapping from file path to file hash."""
    return {filepath.replace('\\', '/'): hash_tree(filepath)
            for filepath in listify(files)}


def process_prop(prop_type: PT, value, build_context):
    """Return a cachable representation of the prop `value` given its type."""
    if prop_type in (PT.Target, PT.TargetList):
        return hashify_targets(value, build_context)
    elif prop_type in (PT.File, PT.FileList):
        return hashify_files(value)
    return value


class Target(types.SimpleNamespace):  # pylint: disable=too-few-public-methods

    _prop_json_blacklist = frozenset((
        'cachable',
        'copy_generated_to',
        'image_caching_behavior',
    ))
    _prop_json_testlist = frozenset((
        'test_flags',
    ))

    def __init__(self, builder_name):
        super().__init__(
            name=None,
            builder_name=builder_name,
            props=Munch(),
            deps=None,
            buildenv=None,
            tags=set(),
            artifacts=ArtifactStore(),
            summary={
                'build_time': None,
                'created': None,
                'accessed': None,
            },
            info={
                'test_time': None,
                'fail_count': 0,
            },
            is_dirty=False,
            tested={},
            _hash=None,
            _json=None,
            _test_json=None,
        )

    def __repr__(self):
        keys = ['name', 'builder_name', 'props', 'deps', 'buildenv', 'tags']
        items = ('{}={!r}'.format(k, self.__dict__[k]) for k in keys)
        return '{}({})'.format(type(self).__name__, ', '.join(items))

    def compute_target_json(self, build_context, prop_blacklist, deps_hashes):
        """Compute a JSON serialization of this target for caching
           purposes.

        The serialization includes:
        - The build flavor
        - The builder name
        - Target tags
        - Hashes of target dependencies & buildenv
        - Processed props (where target props are replaced with their hashes,
          and file props are replaced with mapping from file name to its hash)

        It specifically does NOT include:
        - Artifacts produced by the target

        The target name is currently included, although it would be better off
        to leave it out, and allow targets to be renamed without affecting
        their caching status (if it's just a rename).
        It is currently included because it's the easy way to account for the
        fact that when cached artifacts are restored, their path may be a
        function of the target name in non-essential ways (such as a workspace
        dir name).

        prop_blacklist - props we don't put in the json
        deps_hashes - precalculated hashes of direct dependencies
        """
        props = {}
        for prop in self.props:
            if prop in self._prop_json_blacklist or prop in prop_blacklist:
                continue
            sig_spec = Plugin.builders[self.builder_name].sig.get(prop)
            if sig_spec is None:
                continue
            if prop not in self._prop_json_testlist:
                props[prop] = process_prop(sig_spec.type, self.props[prop],
                                           build_context)
        json_dict = dict(
            # TODO: avoid including the name in the hashed json...
            name=self.name,
            builder_name=self.builder_name,
            deps=sorted(deps_hashes),
            props=props,
            buildenv=hashify_targets(self.buildenv, build_context),
            tags=sorted(list(self.tags)),
            flavor=build_context.conf.flavor,  # TODO: any other conf args?
            # yabt_version=__version__,  # TODO: is this needed?
        )
        if build_context.conf.get('cache_salt', None) is not None:
            json_dict['cache_salt'] = build_context.conf.get('cache_salt',
                                                             None),
        return json.dumps(json_dict, sort_keys=True, indent=4)

    def compute_test_json(self, build_context):
        """
        Compute the json representing the test of this target. it includes only
        the test props.
        """
        test_props = {}
        for prop in self.props:
            if prop in self._prop_json_blacklist:
                continue
            sig_spec = Plugin.builders[self.builder_name].sig.get(prop)
            if sig_spec is None:
                continue
            if prop in self._prop_json_testlist:
                test_props[prop] = process_prop(sig_spec.type,
                                                self.props[prop],
                                                build_context)
        json_test_dict = dict(
            props=test_props,
        )
        return json.dumps(json_test_dict, sort_keys=True, indent=4)

    def compute_json(self, build_context):
        """
        Compute and store all json representing the target.
        """
        builder = Plugin.builders[self.builder_name]
        if builder.cache_json_func:
            self._json = builder.cache_json_func(build_context, self)
        else:
            self._json = self.compute_target_json(
                build_context, [],
                [build_context.targets[target_name].hash(build_context)
                 for target_name in listify(self.deps)])
        self._test_json = self.compute_test_json(build_context)

    def json(self, build_context) -> str:
        """Return JSON serialization of this target for caching purposes."""
        if self._json is None:
            self.compute_json(build_context)
        return self._json

    def test_json(self, build_context) -> str:
        """Return JSON serialization of the test target for caching purposes.
        """
        if self._test_json is None:
            self.compute_json(build_context)
        return self._test_json

    def compute_hash(self, build_context):
        """Compute and store the hash of this target for caching purposes.

        The hash is computed over the target JSON representation.
        """
        m = md5()
        m.update(self.json(build_context).encode('utf8'))
        self._hash = m.hexdigest()
        m = md5()
        m.update(self.test_json(build_context).encode('utf8'))
        self._test_hash = m.hexdigest()

    def hash(self, build_context) -> str:
        """Return the hash of this target for caching purposes."""
        if self._hash is None:
            self.compute_hash(build_context)
        return self._hash

    def test_hash(self, build_context) -> str:
        """Return the hash of this test target for caching purposes."""
        if self._test_hash is None:
            self.compute_hash(build_context)
        return self._test_hash


class ImageCachingBehavior(types.SimpleNamespace):

    def __init__(self, name, tag, behavior: dict = {}):
        super().__init__(
            remote_image='{}:{}'.format(
                behavior.get('remote_image_name', name),
                behavior.get('remote_image_tag', tag)),
            pull_if_not_cached=behavior.get('pull_if_not_cached', False),
            pull_if_cached=behavior.get('pull_if_cached', False),
            allow_build_if_not_cached=behavior.get(
                'allow_build_if_not_cached', True),
            skip_build_if_cached=behavior.get('skip_build_if_cached', False),
            push_image_after_build=behavior.get(
                'push_image_after_build', False),
        )



# -*- coding: utf-8 -*-

# Copyright 2019 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A global cache implemented with google cloud storage
~~~~~~~~~~~~~~~~~~~

:author: Dana Shamir
"""

from google.cloud import storage, exceptions
import os
from os.path import join
from typing import Dict

from .global_cache import GlobalCache, SUMMARY_FILE, ARTIFACTS_FILE, \
    TESTS_FILE, ARTIFACTS_DIR, TARGETS_DIR
from .logging import make_logger


logger = make_logger(__name__)


class GSGlobalCache(GlobalCache):
    def __init__(self, gce_project, bucket, directory=None):
        self.gce_project = gce_project
        self.bucket_name = bucket
        self.targets_dir = join(directory, TARGETS_DIR) if directory else \
            TARGETS_DIR
        self.artifacts_dir = join(directory, ARTIFACTS_DIR) if directory \
            else ARTIFACTS_DIR
        self.storage_client = None
        self.bucket = None

    def _create_client(self):
        if not self.storage_client:
            self.storage_client = storage.Client(self.gce_project)
        if not self.bucket:
            self.bucket = self.storage_client.get_bucket(self.bucket_name)

    def has_cache(self, target_hash: str):
        self._create_client()
        return self.bucket.blob(join(self.targets_dir, target_hash,
                                     SUMMARY_FILE)).exists()

    def download_summary(self, target_hash: str, dst: str) -> bool:
        return self.download_meta_file(target_hash, SUMMARY_FILE, dst)

    def download_artifacts_meta(self, target_hash: str, dst: str) -> bool:
        return self.download_meta_file(target_hash, ARTIFACTS_FILE, dst)

    def download_test_cache(self, target_hash: str, dst: str) -> bool:
        return self.download_meta_file(target_hash, TESTS_FILE, dst)

    def download_meta_file(self, target_hash: str, src: str, dst: str) -> bool:
        self._create_client()
        src_blob = self.bucket.blob(join(self.targets_dir, target_hash, src))
        try:
            src_blob.download_to_filename(dst)
        except exceptions.NotFound:
            # When there is a failure downloading the file from gs, sometimes
            # the dst file is created empty
            try:
                os.remove(dst)
            except FileNotFoundError:
                pass
            return False
        return True

    def download_artifacts(self, artifacts_hashes: Dict[str, int], dst: str):
        self._create_client()
        if artifacts_hashes:
            # TODO(Dana): make this work in batch.
            # see https://github.com/googleapis/google-cloud-python/issues/3139
            for artifact_hash, permissions in artifacts_hashes.items():
                src_blob = self.bucket.blob(join(self.artifacts_dir,
                                                 artifact_hash))
                try:
                    dst_path = join(dst, artifact_hash)
                    src_blob.download_to_filename(dst_path)
                    # This is here because when downloading from gs we don't
                    # get the files with the right permissions.
                    os.chmod(dst_path, permissions)
                except exceptions.NotFound:
                    return False
        return True

    def create_target_cache(self, target_hash: str):
        pass

    def upload_summary(self, target_hash: str, src: str):
        self.upload_target_meta(target_hash, src, SUMMARY_FILE)

    def upload_artifacts_meta(self, target_hash: str, src: str):
        self.upload_target_meta(target_hash, src, ARTIFACTS_FILE)

    def upload_artifacts(self, artifacts_hashes: Dict[str, int], src: str):
        self._create_client()
        # TODO(Dana): make this work in batch
        for artifact_hash in artifacts_hashes.keys():
            self.bucket.blob(join(self.artifacts_dir, artifact_hash))\
                .upload_from_filename(join(src, artifact_hash))

    def upload_test_cache(self, target_hash: str, src: str):
        self.upload_target_meta(target_hash, src, TESTS_FILE)

    def upload_target_meta(self, target_hash, src, dst):
        self._create_client()
        self.bucket.blob(join(self.targets_dir, target_hash, dst))\
            .upload_from_filename(src)



# -*- coding: utf-8 -*-

# Copyright 2016 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt ProtoBuf builder
~~~~~~~~~~~~~~~~~~~~~

:author: Zohar Rimon, Itamar Ostricher
"""


import os
from os.path import dirname, isfile, join, relpath, splitext
from pathlib import Path, PurePath

from ostrich.utils.path import commonpath
from ostrich.utils.text import get_safe_path

from ..artifact import ArtifactType as AT
from ..compat import walk
from ..extend import (
    PropType as PT, register_build_func, register_builder_sig,
    register_manipulate_target_hook)
from ..logging import make_logger
from ..utils import link_files, rmtree, yprint, link_node

logger = make_logger(__name__)


register_builder_sig(
    'Proto',
    [('sources', PT.FileList),
     ('in_buildenv', PT.Target),
     ('proto_cmd', PT.list, 'protoc'),
     ('gen_python', PT.bool, True),
     ('gen_cpp', PT.bool, True),
     ('gen_go', PT.bool, False),
     ('gen_python_rpcz', PT.bool, False),
     ('gen_cpp_rpcz', PT.bool, False),
     ('gen_python_grpc', PT.bool, False),
     ('gen_cpp_grpc', PT.bool, False),
     ('gen_go_grpc', PT.bool, False),
     ('gen_descriptor', PT.bool, False),
     ('grpc_plugin_path', PT.str, '/usr/local/bin/grpc_cpp_plugin'),
     ('copy_generated_to', PT.File, None),
     ('cmd_env', None),
     ])

register_builder_sig('ProtoCollector')


@register_build_func('Proto')
def proto_builder(build_context, target):
    yprint(build_context.conf, 'Build ProtoBuf', target)
    workspace_dir = build_context.get_workspace('ProtoBuilder', target.name)
    # make sure workdir is clean
    rmtree(workspace_dir)
    proto_gen_dir = build_context.conf.get('proto_gen_dir', 'proto')
    proto_dir = join(workspace_dir, proto_gen_dir)
    # Collect proto sources from this target and all dependecies,
    # to link under the target sandbox dir
    protos = [source for source in target.props.sources
              if source.endswith('.proto')]
    for dep in build_context.generate_all_deps(target):
        if 'sources' in dep.props:
            protos.extend(source for source in dep.props.sources
                          if source.endswith('.proto'))
    link_files(protos, proto_dir, None, build_context.conf)
    buildenv_workspace = build_context.conf.host_to_buildenv_path(
        workspace_dir)
    protoc_cmd = target.props.proto_cmd + ['--proto_path', buildenv_workspace]
    descriptor_path = join(proto_gen_dir,
                           get_safe_path(target.name.lstrip(':')) +
                           '_descriptor.pb')
    if target.props.gen_cpp:
        protoc_cmd.extend(('--cpp_out', buildenv_workspace))
    if target.props.gen_python:
        protoc_cmd.extend(('--python_out', buildenv_workspace))
    if target.props.gen_go and not target.props.gen_go_grpc:
        protoc_cmd.extend(('--go_out', buildenv_workspace))
    if target.props.gen_cpp_grpc:
        protoc_cmd.extend(
            ('--grpc_out', buildenv_workspace,
             '--plugin=protoc-gen-grpc={}'
             .format(target.props.grpc_plugin_path)))
    if target.props.gen_python_grpc:
        protoc_cmd.extend(('--grpc_python_out', buildenv_workspace))
    if target.props.gen_go_grpc:
        protoc_cmd.append(
            '--go_out=plugins=grpc:{}'.format(buildenv_workspace))
    if target.props.gen_go or target.props.gen_go_grpc:
        protoc_cmd.append(
            '--go_opt=paths=source_relative'
        )
    if target.props.gen_cpp_rpcz:
        protoc_cmd.extend(('--cpp_rpcz_out', buildenv_workspace))
    if target.props.gen_python_rpcz:
        protoc_cmd.extend(('--python_rpcz_out', buildenv_workspace))
    if target.props.gen_descriptor:
        protoc_cmd.extend(
            ('--include_imports', '--descriptor_set_out',
             (PurePath(buildenv_workspace) / descriptor_path).as_posix()))
    protoc_cmd.extend(
      (PurePath(buildenv_workspace) / proto_gen_dir / src).as_posix()
      for src in target.props.sources)
    build_context.run_in_buildenv(
        target.props.in_buildenv, protoc_cmd, target.props.cmd_env)
    generated_files = []

    def process_generated(gen_file: str, artifact_type: AT):
        if not isfile(gen_file):
            logger.error('Missing expected generated file: {}', gen_file)
            return
        target.artifacts.add(
            artifact_type,
            relpath(gen_file, build_context.conf.project_root),
            relpath(gen_file, workspace_dir))
        generated_files.append(gen_file)

    def create_init_py(path: str):
        init_py_path = join(path, '__init__.py')
        Path(init_py_path).touch(exist_ok=True)
        return init_py_path

    # Add generated files to artifacts / generated list
    for src in target.props.sources:
        src_base = join(proto_dir, splitext(src)[0])
        process_generated(src_base + '.proto', AT.proto)
        if target.props.gen_python:
            process_generated(src_base + '_pb2.py', AT.gen_py)
        if target.props.gen_cpp:
            process_generated(src_base + '.pb.cc', AT.gen_cc)
            process_generated(src_base + '.pb.h', AT.gen_h)
        if target.props.gen_go or target.props.gen_go_grpc:
            process_generated(src_base + '.pb.go', AT.gen_go)
        if target.props.gen_python_rpcz:
            process_generated(src_base + '_rpcz.py', AT.gen_py)
        if target.props.gen_cpp_rpcz:
            process_generated(src_base + '.rpcz.cc', AT.gen_cc)
            process_generated(src_base + '.rpcz.h', AT.gen_h)
        if target.props.gen_python_grpc:
            process_generated(src_base + '_pb2_grpc.py', AT.gen_py)
        if target.props.gen_cpp_grpc:
            process_generated(src_base + '.grpc.pb.cc', AT.gen_cc)
            process_generated(src_base + '.grpc.pb.h', AT.gen_h)
    if target.props.gen_descriptor:
        process_generated(join(workspace_dir, descriptor_path),
                          AT.proto_descriptor)

    # Create __init__.py files in all generated directories with Python files
    if target.props.gen_python or target.props.gen_python_rpcz:
        py_dirs = set(('',))
        for src in target.props.sources:
            py_dir = dirname(src)
            while py_dir not in py_dirs:
                py_dirs.add(py_dir)
                py_dir = dirname(py_dir)
        for py_dir in py_dirs:
            process_generated(create_init_py(join(proto_dir, py_dir)),
                              AT.gen_py)

    # Copy generated files to external destination
    # TODO: remove this? (deprecate copy_generated_to?)
    if target.props.copy_generated_to:
        link_files(generated_files, target.props.copy_generated_to,
                   workspace_dir, build_context.conf)


def add_gen_python_path(target):
    if target.props.gen_python or target.props.gen_python_rpcz:
        join_env = target.props.packaging_params.pop('semicolon_join_env', {})
        if 'PYTHONPATH' in join_env:
            if '/usr/src/gen' not in join_env['PYTHONPATH'].split(':'):
                join_env['PYTHONPATH'] += ':/usr/src/gen'
        else:
            join_env['PYTHONPATH'] = '/usr/src/gen'
        target.props.packaging_params['semicolon_join_env'] = join_env


@register_manipulate_target_hook('Proto')
def proto_manipulate_target(build_context, target):
    target.buildenv = target.props.in_buildenv
    # manipulating the packaging_params (and storing it in the target props)
    # during target extraction (as opposed to build time), because:
    # 1. it should affect the target hash (for cache)
    # 2. it is used by docker builder further down the graph, so it should be
    #    populated if the target is cached (and build func is never called)
    add_gen_python_path(target)


@register_build_func('ProtoCollector')
def proto_collector_builder(build_context, target):
    """
    Collect all the .proto files that this target depend on to a directory
    under yabtwork/*/ProtoCollector.
    If this target depends on a proto target, then all the .proto files needed
    for building this proto will get to this directory.
    TODO(Dana): Support not building the proto (we only need the .proto files)
    """
    workspace_dir = build_context.get_workspace('ProtoCollector', target.name)
    for dep in build_context.generate_all_deps(target):
        artifact_map = dep.artifacts.get(AT.proto)
        if not artifact_map:
            continue
        for dst, src in artifact_map.items():
            target_file = join(workspace_dir, dst)
            link_node(join(build_context.conf.project_root, src), target_file)
            target.artifacts.add(
                AT.proto,
                relpath(target_file, build_context.conf.project_root),
                relpath(target_file, workspace_dir))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2016 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt
~~~~

:author: Itamar Ostricher
"""

import sys

from collections import namedtuple

from . import __oneliner__, __version__
from .buildcontext import BuildContext
from .cli import init_and_get_conf
from .config import Config, BUILD_PROJ_FILE
from .extend import Plugin
from .graph import populate_targets_graph
from .dot import write_dot
from .logging import make_logger
from .target_info import get_target_info_json
from .target_utils import parse_target_selectors, split
from .utils import fatal


YabtCommand = namedtuple('YabtCommand', ['func', 'requires_project'])


def cmd_version(unused_conf):
    """Print out version information about YABT and detected builders."""
    import pkg_resources
    print('This is {} version {}, imported from {}'
          .format(__oneliner__, __version__, __file__))
    if len(Plugin.builders) > 0:
        print('setuptools registered builders:')
    for entry_point in pkg_resources.iter_entry_points('yabt.builders'):
        print('  {0.module_name}.{0.name} (dist {0.dist})'.format(entry_point))


# def cmd_help(unused_conf):
#   parser.print_help()
#   sys.exit(0)


def cmd_list(unused_conf: Config):
    """Print out information on loaded builders and hooks."""
    for name, builder in sorted(Plugin.builders.items()):
        if builder.func:
            print('+- {0:16s} implemented in {1.__module__}.{1.__name__}()'
                  .format(name, builder.func))
        else:
            print('+- {0:16s} loaded with no builder function'.format(name))
        for hook_name, hook_func in sorted(Plugin.get_hooks_for_builder(name)):
            print('  +- {0} hook implemented in '
                  '{1.__module__}.{1.__name__}()'
                  .format(hook_name, hook_func))


def cmd_build(conf: Config, run_tests: bool=False):
    """Build requested targets, and their dependencies."""
    build_context = BuildContext(conf)
    populate_targets_graph(build_context, conf)
    build_context.build_graph(run_tests=run_tests)
    build_context.write_artifacts_metadata()


def cmd_test(conf: Config):
    """Build requested targets and their dependencies and run test nodes."""
    # TODO: Automatic test discovery?
    cmd_build(conf, run_tests=True)


def cmd_dot(conf: Config):
    """Print out a neat targets dependency tree based on requested targets.

    Use graphviz to render the dot file, e.g.:

    > ybt dot :foo :bar | dot -Tpng -o graph.png
    """
    build_context = BuildContext(conf)
    populate_targets_graph(build_context, conf)
    if conf.output_dot_file is None:
        write_dot(build_context, conf, sys.stdout)
    else:
        with open(conf.output_dot_file, 'w') as out_file:
            write_dot(build_context, conf, out_file)


def cmd_info(conf: Config):
    build_context = BuildContext(conf)
    populate_targets_graph(build_context, conf)
    print(get_target_info_json(conf, build_context))


def cmd_tree(conf: Config):
    """Print out a neat targets dependency tree based on requested targets."""
    build_context = BuildContext(conf)
    populate_targets_graph(build_context, conf)

    def print_target_with_deps(target, depth=2):
        print('{: >{}}{}'.format('+-', depth, target.name))
        for dep in sorted(
                build_context.target_graph.neighbors(target.name)):
            print_target_with_deps(build_context.targets[dep], depth + 2)

    if conf.targets:
        for target_name in sorted(parse_target_selectors(conf.targets, conf)):
            mod, name = split(target_name)
            if name == '*':
                for target_name in sorted(
                        build_context.targets_by_module[mod]):
                    print_target_with_deps(build_context.targets[target_name])
            else:
                print_target_with_deps(build_context.targets[target_name])
    else:
        for _, target in sorted(build_context.targets.items()):
            print_target_with_deps(target)


def main():
    """Main `ybt` console script entry point - run YABT from command-line."""
    conf = init_and_get_conf()
    logger = make_logger(__name__)
    logger.info('YaBT version {}', __version__)
    handlers = {
        'build': YabtCommand(func=cmd_build, requires_project=True),
        'dot': YabtCommand(func=cmd_dot, requires_project=True),
        'test': YabtCommand(func=cmd_test, requires_project=True),
        'tree': YabtCommand(func=cmd_tree, requires_project=True),
        'info': YabtCommand(func=cmd_info, requires_project=True),
        'version': YabtCommand(func=cmd_version, requires_project=False),
        'list-builders': YabtCommand(func=cmd_list, requires_project=False),
    }
    command = handlers[conf.cmd]
    if command.requires_project and not conf.in_yabt_project():
        fatal('Not a YABT project (or any of the parent directories): {}',
              BUILD_PROJ_FILE)
    try:
        command.func(conf)
    except Exception as ex:
        fatal('{}', ex)


if __name__ == '__main__':
    main()
